import numpy as np
import json
from onnxruntime import InferenceSession
from transformers import AutoTokenizer

class MorphPredictor:
    def __init__(self) -> None:
        pass

    def load(self, path, device="CPU"):
        self.session = InferenceSession(f"{path}/model.onnx", providers=["CUDAExecutionProvider" if device == "CUDA" else "CPUExecutionProvider"])

        with open(f"{path}/config.json", "r") as f:
            self.id2label = json.load(f)["id2label"]
        self.tokenizer = AutoTokenizer.from_pretrained(path)
        #self.tokenizer.model_input_names = ["input_ids", "attention_mask"]

    def collect_pre_entities(self, sentence, input_ids, scores, offset_mapping, special_tokens_mask):
        pre_entities = []
        for idx, token_scores in enumerate(scores):
            if special_tokens_mask[idx]:
                continue
            word = self.tokenizer.convert_ids_to_tokens(int(input_ids[idx]))
           
            if offset_mapping is not None:
                start_ind, end_ind = offset_mapping[idx]
                if not isinstance(start_ind, int):
                    start_ind = int(start_ind)
                    end_ind = int(end_ind)
                word_ref = sentence[start_ind:end_ind]
                if getattr(self.tokenizer._tokenizer.model, "continuing_subword_prefix", None):
                    is_subword = len(word) != len(word_ref)
                else:
                    is_subword = sentence[start_ind - 1 : start_ind] != " " if start_ind > 0 else False
    
                if int(input_ids[idx]) == self.tokenizer.unk_token_id:
                    word = word_ref
                    is_subword = False
            else:
                start_ind = None
                end_ind = None
                is_subword = False
    
            pre_entity = {
                "word": word,
                "scores": token_scores,
                "start": start_ind,
                "end": end_ind,
                "index": idx,
                "is_subword": is_subword,
            }
            pre_entities.append(pre_entity)
        return pre_entities
    
    def aggregate_word(self, entities, aggregation_strategy):
        word = self.tokenizer.convert_tokens_to_string([entity["word"] for entity in entities])
        if aggregation_strategy == "FIRST":
            scores = entities[0]["scores"]
            idx = scores.argmax()
            score = scores[idx]
            entity = self.id2label[str(idx)]
        elif aggregation_strategy == "MAX":
            max_entity = max(entities, key=lambda entity: entity["scores"].max())
            scores = max_entity["scores"]
            idx = scores.argmax()
            score = scores[idx]
            entity = self.id2label[str(idx)]
        elif aggregation_strategy == "AVERAGE":
            scores = np.stack([entity["scores"] for entity in entities])
            average_scores = np.nanmean(scores, axis=0)
            entity_idx = average_scores.argmax()
            entity = self.id2label[str(entity_idx)]
            score = average_scores[entity_idx]
    
        new_entity = {
            "entity": entity,
            "score": score,
            "word": word,
            "start": entities[0]["start"],
            "end": entities[-1]["end"],
        }
        return new_entity
    
    def aggregate_words(self, entities, aggregation_strategy):
        word_entities = []
        word_group = None
        for entity in entities:
            if word_group is None:
                word_group = [entity]
            elif entity["is_subword"]:
                word_group.append(entity)
            else:
                word_entities.append(self.aggregate_word(word_group, aggregation_strategy))
                word_group = [entity]
        word_entities.append(self.aggregate_word(word_group, aggregation_strategy))
        return word_entities
    
    def predict_morph(self, text):
        inputs = self.tokenizer(text, return_offsets_mapping=True, return_special_tokens_mask=True, return_tensors="np")
        offset_mapping, special_tokens_mask, input_ids = inputs.pop('offset_mapping')[0], inputs.pop('special_tokens_mask')[0], inputs['input_ids'][0]
        inputs = {k: v.astype(np.int64) for k, v in inputs.items()}
        outputs = self.session.run(None, inputs)
        logits = outputs[0]
        maxes = np.max(logits, axis=-1, keepdims=True)
        shifted_exp = np.exp(logits - maxes)
        scores = shifted_exp / shifted_exp.sum(axis=-1, keepdims=True)
        pre_entities = self.collect_pre_entities(text, input_ids, scores[0], offset_mapping, special_tokens_mask)
        grouped_entities = self.aggregate_words(pre_entities, "AVERAGE")
        return grouped_entities


import json
import re
import gzip
from os.path import join as join_path


class RuleEngine:
    def __init__(self):
        self.wordforms = {}
        self.forms_dict = {}

    def load(self, path, morph):
        self.morph = morph
        with open(file=join_path(path, "accents.json"), mode='r') as f:
            self.wordforms = json.load(f)
        with open(file=join_path(path, "forms.json"), mode='r') as f:
            self.forms_dict = json.load(f)
        
    def split_by_words(self, string):
        string = string.replace(" - ",' ~ ')
        result = re.findall(r"\w*(?:\+\w+)*|[^\w\s]+", string.lower())
        return [res for res in result if res]

    def all_elements_equal(self, lst):
        return all(x == lst[0] for x in lst)

    def calculate_similarity(self, set1, set2):
        intersection = set1 & set2
        union = set1 | set2
        return (len(intersection) / len(union)) * 100 if union else 100

    def check_lemmas(self, word):
        lemmas = [i['lemma'] for i in word["interpretations"]]
        if len(lemmas) > 2 or not self.all_elements_equal(lemmas):
            for interpretation in word['interpretations']:
                if interpretation['lemma'] == word['lemma']:
                    return interpretation["accentuated"]
        return None

    def compatible(self, interpretation, tag):
        forms = " ".join(self.forms_dict.get(form, "") for form in interpretation.split())
        tags = set(tag.replace("Voice=Act", "").replace("VerbForm=Fin", "").split("|"))
        return self.calculate_similarity(tags, set(forms.split()))

    def accentuate_word(self, word):
        accented_lemma = self.check_lemmas(word)
        if accented_lemma:
            return accented_lemma
        else:
            check_tag = [self.compatible(interpretation['form'], word['tag'])
                         for interpretation in word['interpretations']]
            max_compatible_index = check_tag.index(max(check_tag))
            return word["interpretations"][max_compatible_index]["accentuated"]

    def tokenize(self, text):
        text_tokenized = self.split_by_words(text)
        if any(word in self.wordforms for word in text_tokenized):
            processed = self.morph.process(text)
            return [
                {"token": word, "tag": tags, "homograph": word.lower() in self.wordforms, "lemma": lemma, "interpretations": self.wordforms.get(word.lower(), [])}
                for word, tags, lemma in processed
            ]
        else:
            return [{"token": token, "homograph": False} for token in text_tokenized]

    def accentuate(self, text):
        words = self.tokenize(text)
        return [self.accentuate_word(word) if word["homograph"] else word['token'] for word in words]
